import socket
import os
import os.path
import typing
import shlex

class Key(dict):
    def from_line(line):
        key = Key()
        for token in shlex.split(line):
            if '=' in token:
                k, v = token.split('=')
                key[k] = v
            elif token.endswith('!'):
                key[token] = None
        return key
    
    def has_secret(self):
        return any(map(lambda kv: kv[0].endswith('!'), self.items()))

    def to_query(self):
        return dict(filter(lambda kv: kv[1] is not None, self.items()))

    # this is silly but apparently according to the freedesktop shitspec there can only be one,
    # so return whichever is first
    def get_secret(self):
        return next(filter(lambda kv: kv[0].endswith('!'), self.items()))[1]

class Client:
    def __init__(self, addr = None):
        if not addr:
            addr = os.path.join(os.getenv('XDG_RUNTIME_DIR'), 'himitsu')

        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(addr)

    def query(self, **kwargs):
        return self._query(kwargs)

    def query_decrypt(self, **kwargs):
        return self._query(kwargs, True)

    def _query(self, query: typing.Dict[str, str], decrypt: bool = False):
        params = ['query']

        if decrypt:
            params.append('-d')

        for arg in query:
            params.append(f'{arg}={query[arg]}')

        qs = ' '.join(params) + '\n'

        self.sock.send(qs.encode('utf-8'))
        raw_res = self._read_raw_response()

        res = []
        for line in raw_res:
            if line.startswith('error '):
                raise Exception(line[6:])
            if not line.startswith('key '):
                raise Exception(f'parse error: {line}')
            line = line[4:]
            # parse 'key ...' line
            key = Key.from_line(line)
            res.append(key)
        return res

    def _read_raw_response(self):
        res = ''
        while not '\nend\n' in res:
            res += self.sock.recv(4096).decode('utf-8')
            if res.startswith('end\n'):
                break
        res = res.strip()
        res_lines = res.splitlines()
        res_lines.pop()
        return res_lines
